// $Id: testRCV.java,v 1.1 2005/11/30 20:46:33 cssharp Exp $


/**
 * PC Application for Telos Mote Trilateration
 *
 * @author Gregory Parsons <gsparson@nc.rr.com>
 * @since  0.1
 * @CSC714 Fall 2005
 */

import net.tinyos.message.*;
import net.tinyos.util.*;
import java.lang.String;

import java.io.*; 
import java.text.*;

public class testRCV implements MessageListener
{
  static final short TOS_UART_ADDR = 0x007e;
  private String m_strings[];
  private MoteIF m_moteif;
  private int m_nSend;
  private int Received;
  private int[][] step1 = new int[10][10];	// raw RSSI values
  private int[][] step2 = new int[10][10];	// averaged RSSI values
  private int[][] step3 = new int[10][10];	// inverted RSSI values
  private double[][] step4 = new double[10][10];// distance between all motes as measured in inches
  private double[] X = new double[10];		// X-position of all motes
  private double[] Y = new double[10];		// Y-position of all motes
  private double[] Z = new double[10];		// Z-position of all motes
  private int display = 0;			// denotes the display mode
  private int MaxID = 0;			// represents the number of motes engaged in network communication
  private int cycle = 0;			// the last cycle number received; these values are generated by the attached mote

  testRCV( String[] args )
  {
    /*if( args.length <= 0 )
    {
      System.out.println("usage: testRCV [string]");
      System.exit(0);
    }*/

    try
    {
      m_moteif = new MoteIF((Messenger)null);
      m_moteif.registerListener(new RCVMsg(), this);
    }
    catch (Exception e)
    {
      System.out.println("ERROR: Couldn't contact serial forwarder.");
      System.exit(1);
    }

    m_strings = args;
    m_nSend = -1;
    m_moteif.start();
  }

  public synchronized void send( String str )
  {
    try
    {
      RCVMsg m = new RCVMsg();
      m.setString_str( str );
      m_moteif.send( MoteIF.TOS_BCAST_ADDR, m );
      System.out.println( "Send> " + m.getString_str() );
    }
    catch (IOException e)
    {
      e.printStackTrace();
      System.out.println("ERROR: Can't send message");
      System.exit(1);
    }
  }

  public boolean sendNextString()
  {
    if( ++m_nSend >= m_strings.length )
    {
      m_nSend = 0;
      return false;
    }
    //Received = 0;
	// determine the number of packets to be requested and then waited for
    Received = 1; //Integer.parseInt(m_strings[m_nSend]);

    send( m_strings[m_nSend] );
    return true;
  }

  public void trilateration(int step)
  // The principles and derivations of this function come almost exclusively from http://en.wikipedia.org/wiki/Trilateration
  // This function utilizes the distance information found in 'step4' to mathematically derive each motes relative position in a 3D space
  // if 'step' is equal to 6 then output will halt whenever an answer was not found
  {
    int mote;
    int success = 1;
    double d, i, j, local, r1, r2, r3;
    DecimalFormat twoDec = new DecimalFormat("0.00");
    d = i = j = 0;
    for(mote = 0; mote <= MaxID; mote++) {	// reset all mote positions
      X[mote] = 0.0f;
      Y[mote] = 0.0f;
      Z[mote] = 0.0f;
    }
    if(MaxID > 0) d = X[1] = step4[0][1];	// derive the position of base-station #2, mote 1, to fix the X-Axis
    if(MaxID > 1) {
      i = X[2] = (step4[0][2]*step4[0][2] - step4[1][2]*step4[1][2] + d*d) / (2 * d);	// derive the relative position of base-station #3, mote 2, to fix the Z-Plane
      local = step4[0][2]*step4[0][2] - i * i;
      if(local < 0) success = 0;	// if 'local' is negative then no answer exists, set success equal to false
      j = Y[2] = Math.sqrt(local);
    }
    for(mote = 3; mote <= MaxID; mote++) {	// derive the position of each target mote utilizing the three previously positioned base stations
      r1 = step4[0][mote] * step4[0][mote];
      r2 = step4[1][mote] * step4[1][mote];
      r3 = step4[2][mote] * step4[2][mote];
      local = r1 - r2 + d*d;
      X[mote] = local / (2*d);
      Y[mote] = ((r1 - r3 + (X[mote]-i)*(X[mote]-i)) / (2*j)) + (j/2);
      Y[mote] -= (local*local) / (8*d*d*j);
      local = r1 - X[mote]*X[mote] - Y[mote]*Y[mote];
      if(local < 0) success = 0;	// if 'local' is negative then no answer exists, set success equal to false
      Z[mote] = Math.sqrt(local);
    }
    if((step == 5) || (success == 1)) {	// if mathematical derivation succeeded or 'step' is equal to 5 then print all relative positions
      for(mote = 0; mote <= MaxID; mote++)
        System.out.println(mote + "(" + twoDec.format(X[mote]) + "," + twoDec.format(Y[mote]) + "," + twoDec.format(Z[mote]) + ")");
      System.out.println("** " + cycle + " MaxID: " + MaxID + " **");
    }
  }

  public void reDisplay(int step)
  // this function prints a pretty matrix of size 'MaxID' x 'MaxID' containing one of the four preparation matrix
  // 1: prints the raw packet data
  // 2: prints the averaged packet data matrix
  // 3: prints the averaged/negated data matrix
  // 4: prints the calculated channel distance matrix
  {
    int row = 0;
    int col = 0;
    DecimalFormat twoDecPlcFormatter = new DecimalFormat("0.00");
    for(row = 0; row <= MaxID; row++) {
      System.out.print(row + "[");
      for(col = 0; col <= MaxID; col++) {
	if(step == 1) System.out.print(step1[row][col] + ",");
	if(step == 2) System.out.print(step2[row][col] + ",");
	if(step == 3) System.out.print(step3[row][col] + ",");
	if(step >= 4) System.out.print(twoDecPlcFormatter.format(step4[row][col]) + ",");
      }
      System.out.println("]");
    }
    System.out.println("\t MaxID=" + MaxID + ", cyc=" + cycle);
  }

  public void reCompute()
  // This function begins the long process of codifying and analysing the available data
  // The primary function completed in this function is to average the RSSI register values collected
  // in matrix 'step1' storing the answers in matrix 'step2'
  // these channel averaged values are then translated into feet and stored into matrix 'step4'
  {
    int row = 0;
    int col = 0;
    int reading;
    for(row = 0; row <= MaxID; row++) {
      for(col = 0; col <= MaxID; col++) {
        reading = step2[row][col] = (step1[row][col] + step1[col][row]) / 2;
	step3[row][col] = (step2[row][col] != 0)?(280 - step2[row][col]):0;
	if(reading >= 253) step4[row][col] = 31.07865f - 0.11236f * reading;
	else step4[row][col] = 183.64f - 0.7137f * reading;
	if(row == col) step4[row][col] = 0.00f;
      }
    }
    // the next step is to print the results depending on which display mode is enabled
    if(display >= 5) trilateration(display);
    if((display >= 1) && (display <= 4)) reDisplay(display);
  }

  public void ParsePacket(int[] packet)
  // This function loads the first step1 matrix by parsing the latest received packet
  // 'packet' is an array of 20 characters consisting of the latest packet to be received
  {
    int row = packet[0];
    int col = 10;
    if((packet[1] <= 10) && (packet[1] != MaxID)) MaxID = packet[1];	// a new MaxID has been detected
    if(packet[2] != cycle) cycle = packet[2];
    for(col = 0; col < 10; col++) {		// transfer the raw packet data to the step1 matrix
      if((packet[col + 10] < 60) && (row != col))
        step1[row][col] = packet[col + 10] + 256;
      else step1[row][col] = packet[col + 10];
    }
    if(packet[0] == MaxID) reCompute();	// if the latest packet is originating from Base-Station #1 then begin analysing the data
  }

  synchronized public void messageReceived( int destAddr, Message m )
  // a message from an attached mote will cause this function to execute
  // This function captures the received packet then calls the parse function
  {
    int n = 0;
    int[] value = new int[20];
    if(display == 0) System.out.print( "Recv> ");
    for(n = 0; n < 20; n++) {
	value[n] = (((RCVMsg)m).getElement_str(n) & 0xff);
        if(display == 0) System.out.print(value[n] + ",");
    }
    if(display == 0) System.out.println("<end>");
    if(display != 0) ParsePacket(value);
  }

  public static void main(String[] args)
  {
    String kb;
    char k = '-';
    int done = 0;
    testRCV m = new testRCV( args );
    m.sendNextString();
    
    while(done != 1) {
      try {
      k = (char)(System.in.read());	// read one character from the PC Keyboard
      } catch (IOException ioe) {
        System.out.println( "IO error:" + ioe );
      } 
      if((k >= '0') && (k <= '8')) m.display = k - '0';	// Number keys change the current display mode
      if((k == 'q') || ( k == 's') || (k == 'h')) {
        System.out.println("cmd " + k);
        if( k == 'q' )		// the 'q' key causes this application to exit
        {
          System.out.println("... done.");
          System.exit(0);
        }
        if( k == 's' ) m.send( "0" );	// the 's' key sends the command to whatever mote is attached to the PC to begin network construction
        if( k == 'h' ) m.send( "99" );	// the 'h' key sends the command to halt whatever network nodes exist
        k = '-';
      }
    }
  }


}

